import os
import unittest

import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.example import example_scheduler
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.utils import restore_scheduler, run_info_table, season_calc
from rubin_scheduler.utils import survey_start_mjd


class TestUtils(unittest.TestCase):
    @unittest.skipUnless(
        os.path.isfile(os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")),
        "Test data not available.",
    )
    def test_nside(self):
        """Test the example scheduler can be set to different nsides."""
        mjd_start = survey_start_mjd()
        _ = example_scheduler(mjd_start=mjd_start, nside=64)
        _ = example_scheduler(mjd_start=mjd_start, nside=8)

    @unittest.skipUnless(
        os.path.isfile(os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")),
        "Test data not available.",
    )
    def test_restore(self):
        """Test we can restore a scheduler properly"""
        # MJD set so it's in test data range
        mjd_start = survey_start_mjd()
        n_visit_limit = 3000

        scheduler = example_scheduler(mjd_start=mjd_start)

        mo = ModelObservatory(mjd_start=mjd_start)
        # Never load too many nights of sky
        mo.sky_model.load_length = 10.0
        mo, scheduler, observations = sim_runner(
            mo,
            scheduler,
            survey_length=30.0,
            verbose=False,
            filename=None,
            n_visit_limit=n_visit_limit,
        )

        # Won't be exact if we restart in the middle of a blob sequence
        # since the queue isn't reconstructed.
        # Also, any scripted observations that get generated
        # during the night (e.g., long gaps observations) will get lost,
        # so need to restart on a new night to ensure identical results.

        nd = np.zeros(observations.size)
        nd[1:] = np.diff(observations["night"])

        break_indx = np.min(np.where((observations["ID"] >= n_visit_limit / 2.0) & (nd != 0))[0])
        new_n_limit = n_visit_limit - break_indx

        new_mo = ModelObservatory(mjd_start=mjd_start)
        # Never load too much sky
        new_mo.sky_model.load_length = 10.0
        new_sched = example_scheduler(mjd_start=mjd_start)

        # Restore some of the observations
        new_sched, new_mo = restore_scheduler(break_indx - 1, new_sched, new_mo, observations, fast=False)

        # Simulate ahead and confirm that it behaves the same
        # as running straight through
        new_mo, new_sched, new_obs = sim_runner(
            new_mo,
            new_sched,
            survey_length=20.0,
            verbose=False,
            filename=None,
            n_visit_limit=new_n_limit,
        )

        # Check that observations taken after restart match those from before
        # Jenkins can be bad at comparing things, so if it thinks
        # they aren't the same, check column-by-column to double check
        if not np.all(new_obs == observations[break_indx:]):
            names = new_obs.dtype.names
            for name in names:
                # If it's a string
                if new_obs[name].dtype == "<U40":
                    assert np.all(new_obs[name] == observations[break_indx:][name])
                # Otherwise should be number-like
                else:
                    assert np.allclose(new_obs[name], observations[break_indx:][name])
        # Didn't need to go by column, the observations after restart
        # match the ones that were taken all at once.
        else:
            assert np.all(new_obs == observations[break_indx:])

        # And again, but this time using the fast array restore
        new_mo = ModelObservatory(mjd_start=mjd_start)
        new_mo.sky_model.load_length = 10.0
        new_sched = example_scheduler(mjd_start=mjd_start)
        new_sched, new_mo = restore_scheduler(break_indx - 1, new_sched, new_mo, observations, fast=True)
        # Simulate ahead and confirm that it behaves the same as
        # running straight through
        new_mo, new_sched, new_obs_fast = sim_runner(
            new_mo,
            new_sched,
            survey_length=20.0,
            verbose=False,
            filename=None,
            n_visit_limit=new_n_limit,
        )

        # Check that observations taken after restart match those from before
        # Jenkins can be bad at comparing things, so if it thinks
        # they aren't the same, check column-by-column to double check
        if not np.all(new_obs_fast == observations[break_indx:]):
            names = new_obs_fast.dtype.names
            for name in names:
                # If it's a string
                if new_obs_fast[name].dtype == "<U40":
                    assert np.all(new_obs_fast[name] == observations[break_indx:][name])
                # Otherwise should be number-like
                else:
                    assert np.allclose(new_obs_fast[name], observations[break_indx:][name])
        # Didn't need to go by column, the observations after restart
        # match the ones that were taken all at once.
        else:
            assert np.all(new_obs_fast == observations[break_indx:])

    def test_season(self):
        """
        Test that the season utils work as intended
        """
        night = 365.25 * 3.5
        plain = season_calc(night)
        assert plain == 3

        mod2 = season_calc(night, modulo=2)
        assert mod2 == 1

        mod3 = season_calc(night, modulo=3)
        assert mod3 == 0

        mod3 = season_calc(night, modulo=3, max_season=2)
        assert mod3 == -1

        mod3 = season_calc(night, modulo=3, max_season=2, offset=-365.25 * 2)
        assert mod3 == 1

        mod3 = season_calc(night, modulo=3, max_season=2, offset=-365.25 * 10)
        assert mod3 == -1

        mod3 = season_calc(night, modulo=3, offset=-365.25 * 10)
        assert mod3 == -1

    def test_run_info_table(self):
        """Test run_info_table gets information"""
        observatory = ModelObservatory(nside=8, mjd_start=survey_start_mjd())
        version_info = run_info_table(observatory)
        # Make a minimal set of keys that ought to be in the info table
        # Update these if the value they're stored as changes
        # (either from run_info_table or observatory.info)
        need_keys = [
            "rubin_scheduler.__version__",
            "hostname",
            "Date, ymd",
            "site_models",
            "skybrightness_pre",
        ]
        have_keys = list(version_info["Parameter"])
        for k in need_keys:
            self.assertTrue(k in have_keys)


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "thetaphi2xyz",
    "even_points",
    "elec_potential",
    "ang_potential",
    "fib_sphere_grid",
    "iterate_potential_random",
    "iterate_potential_smart",
    "even_points_xyz",
    "elec_potential_xyz",
    "xyz2thetaphi",
)

import numpy as np
from scipy.optimize import minimize

from rubin_scheduler.utils import _angular_separation


def thetaphi2xyz(theta, phi):
    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(phi)
    return x, y, z


def xyz2thetaphi(x, y, z):
    phi = np.arccos(z)
    theta = np.arctan2(y, x)
    return theta, phi


def elec_potential(x0):
    """
    Compute the potential energy for electrons on a sphere

    Parameters
    ----------
    x0 : array
       First half of x0 or theta values, secnd half phi

    Returns
    -------
    Potential energy
    """

    theta = x0[0 : int(x0.size / 2)]
    phi = x0[int(x0.size / 2) :]

    x, y, z = thetaphi2xyz(theta, phi)
    # Distance squared
    dsq = 0.0

    indices = np.triu_indices(x.size, k=1)

    for coord in [x, y, z]:
        coord_i = np.tile(coord, (coord.size, 1))
        coord_j = coord_i.T
        d = (coord_i[indices] - coord_j[indices]) ** 2
        dsq += d

    U = np.sum(1.0 / np.sqrt(dsq))
    return U


def potential_single(coord0, x, y, z):
    """
    Find the potential contribution from a single point.
    """

    x0 = coord0[0]
    y0 = coord0[1]
    z0 = coord0[2]
    # Enforce point has to be on a sphere
    rsq = x0**2 + y0**2 + z0**2
    r = np.sqrt(rsq)
    x0 = x0 / r
    y0 = y0 / r
    z0 = z0 / r

    dsq = (x - x0) ** 2 + (y - y0) ** 2 + (z - z0) ** 2
    U = np.sum(1.0 / np.sqrt(dsq))
    return U


def xyz2_u(x, y, z):
    """
    compute the potential
    """
    dsq = 0.0

    indices = np.triu_indices(x.size, k=1)

    for coord in [x, y, z]:
        coord_i = np.tile(coord, (coord.size, 1))
        coord_j = coord_i.T
        dsq += (coord_i[indices] - coord_j[indices]) ** 2

    d = np.sqrt(dsq)
    U = np.sum(1.0 / d)
    return U


def iterate_potential_smart(x0, stepfrac=0.1):
    """
    Calculate the change in potential by shifting points in theta and phi directions
    # wow, that sure didn't work at all.
    """

    theta = x0[0 : x0.size / 2]
    phi = x0[x0.size / 2 :]
    x, y, z = thetaphi2xyz(theta, phi)
    u_input = xyz2_u(x, y, z)

    # Now to loop over each point, and find where it's potenital minimum would be, and move it
    # half-way there.
    xyz_new = np.zeros((x.size, 3), dtype=float)
    mask = np.ones(x.size, dtype=bool)
    for i in np.arange(x.size):
        mask[i] = 0
        fit = minimize(potential_single, [x[i], y[i], z[i]], args=(x[mask], y[mask], z[mask]))
        mask[i] = 1
        xyz_new[i] = fit.x / np.sqrt(np.sum(fit.x**2))

    xyz_input = np.array((x, y, z)).T
    diff = xyz_input - xyz_new

    # Move half way in x-y-z space
    xyz_out = xyz_input + stepfrac * diff
    # Project back onto sphere
    xyz_out = xyz_out.T / np.sqrt(np.sum(xyz_out**2, axis=1))
    u_new = xyz2_u(xyz_out[0, :], xyz_out[1, :], xyz_out[2, :])
    theta, phi = xyz2thetaphi(xyz_out[0, :], xyz_out[1, :], xyz_out[2, :])
    return np.concatenate((theta, phi)), u_new


def iterate_potential_random(x0, stepsize=0.05):
    """
    Given a bunch of theta,phi values, shift things around to minimize potential
    """

    theta = x0[0 : int(x0.size / 2)]
    phi = x0[int(x0.size / 2) :]

    x, y, z = thetaphi2xyz(theta, phi)
    # Distance squared
    dsq = 0.0

    indices = np.triu_indices(x.size, k=1)

    for coord in [x, y, z]:
        coord_i = np.tile(coord, (coord.size, 1))
        coord_j = coord_i.T
        d = (coord_i[indices] - coord_j[indices]) ** 2
        dsq += d

    d = np.sqrt(dsq)

    u_input = 1.0 / d

    # offset everything by a random ammount
    x_new = x + np.random.random(theta.size) * stepsize
    y_new = y + np.random.random(theta.size) * stepsize
    z_new = z + np.random.random(theta.size) * stepsize

    r = (x_new**2 + y_new**2 + z_new**2) ** 0.5
    # put back on the sphere
    x_new = x_new / r
    y_new = y_new / r
    z_new = z_new / r

    dsq_new = 0
    for coord, coord_new in zip([x, y, z], [x_new, y_new, z_new]):
        coord_i_new = np.tile(coord_new, (coord_new.size, 1))
        coord_j = coord_i_new.T
        d_new = (coord_i_new[indices] - coord_j[indices]) ** 2
        dsq_new += d_new
    u_new = 1.0 / np.sqrt(dsq_new)

    u_diff = np.sum(u_new) - np.sum(u_input)
    if u_diff > 0:
        return x0, 0.0
    else:
        theta, phi = xyz2thetaphi(x_new, y_new, z_new)
        return np.concatenate((theta, phi)), u_diff


def ang_potential(x0):
    """
    If distance is computed along sphere rather than through 3-space.
    """
    theta = x0[0 : int(x0.size / 2)]
    phi = np.pi / 2 - x0[int(x0.size / 2) :]

    indices = np.triu_indices(theta.size, k=1)

    theta_i = np.tile(theta, (theta.size, 1))
    theta_j = theta_i.T
    phi_i = np.tile(phi, (phi.size, 1))
    phi_j = phi_i.T
    d = _angular_separation(theta_i[indices], phi_i[indices], theta_j[indices], phi_j[indices])
    U = np.sum(1.0 / d)
    return U


def fib_sphere_grid(npoints):
    """
    Use a Fibonacci spiral to distribute points uniformly on a sphere.

    based on https://people.sc.fsu.edu/~jburkardt/py_src/sphere_fibonacci_grid/sphere_fibonacci_grid_points.py

    Returns theta and phi in radians
    """

    phi = (1.0 + np.sqrt(5.0)) / 2.0

    i = np.arange(npoints, dtype=float)
    i2 = 2 * i - (npoints - 1)
    theta = (2.0 * np.pi * i2 / phi) % (2.0 * np.pi)
    sphi = i2 / npoints
    phi = np.arccos(sphi)
    return theta, phi


def even_points(npts, use_fib_init=True, method="CG", potential_func=elec_potential, maxiter=None):
    """
    Distribute npts over a sphere and minimize their potential, making them
    "evenly" distributed

    Starting with the Fibonacci spiral speeds things up by ~factor of 2.
    """

    if use_fib_init:
        # Start with fibonacci spiral guess
        theta, phi = fib_sphere_grid(npts)
    else:
        # Random on a sphere
        theta = np.random.rand(npts) * np.pi * 2.0
        phi = np.arccos(2.0 * np.random.rand(npts) - 1.0)

    x = np.concatenate((theta, phi))
    # XXX--need to check if this is the best minimizer
    min_fit = minimize(potential_func, x, method="CG", options={"maxiter": maxiter})

    x = min_fit.x
    theta = x[0 : int(x.size / 2)]
    phi = x[int(x.size / 2) :]
    # Looks like I get the same energy values as https://en.wikipedia.org/wiki/Thomson_problem
    return theta, phi


def elec_potential_xyz(x0):
    x0 = x0.reshape(3, int(x0.size / 3))
    x = x0[0, :]
    y = x0[1, :]
    z = x0[2, :]
    dsq = 0.0

    r = np.sqrt(x**2 + y**2 + z**2)
    x = x / r
    y = y / r
    z = z / r
    indices = np.triu_indices(x.size, k=1)

    for coord in [x, y, z]:
        coord_i = np.tile(coord, (coord.size, 1))
        coord_j = coord_i.T
        d = (coord_i[indices] - coord_j[indices]) ** 2
        dsq += d

    U = np.sum(1.0 / np.sqrt(dsq))
    return U


def elec_p_xyx_loop(x0):
    """do this with a brutal loop that can be numba ified"""
    x0 = x0.reshape(3, int(x0.size / 3))
    x = x0[0, :]
    y = x0[1, :]
    z = x0[2, :]
    U = 0.0

    r = np.sqrt(x**2 + y**2 + z**2)
    x = x / r
    y = y / r
    z = z / r

    npts = x.size
    for i in range(npts - 1):
        for j in range(i + 1, npts):
            dsq = (x[i] - x[j]) ** 2 + (y[i] - y[j]) ** 2 + (z[i] - z[j]) ** 2
            U += 1.0 / np.sqrt(dsq)
    return U


def x02sphere(x0):
    x0 = x0.reshape(3, int(x0.size / 3))
    x = x0[0, :]
    y = x0[1, :]
    z = x0[2, :]

    r = np.sqrt(x**2 + y**2 + z**2)
    x = x / r
    y = y / r
    z = z / r

    return np.concatenate((x, y, z))


def even_points_xyz(
    npts,
    use_fib_init=True,
    method="CG",
    potential_func=elec_p_xyx_loop,
    maxiter=None,
    callback=None,
    verbose=True,
):
    """
    Distribute npts over a sphere and minimize their potential, making them
    "evenly" distributed

    Starting with the Fibonacci spiral speeds things up by ~factor of 2.
    """

    if use_fib_init:
        # Start with fibonacci spiral guess
        theta, phi = fib_sphere_grid(npts)
    else:
        # Random on a sphere
        theta = np.random.rand(npts) * np.pi * 2.0
        phi = np.arccos(2.0 * np.random.rand(npts) - 1.0)

    x = np.concatenate(thetaphi2xyz(theta, phi))

    if verbose:
        print("initial potential=", elec_potential_xyz(x))
    # XXX--need to check if this is the best minimizer
    min_fit = minimize(potential_func, x, method="CG", options={"maxiter": maxiter}, callback=callback)

    if verbose:
        print("final potential=", elec_potential_xyz(min_fit.x))
        print("niteration=", min_fit.nit)

    x = x02sphere(min_fit.x)

    # Looks like I get the same energy values as https://en.wikipedia.org/wiki/Thomson_problem
    return x



import numpy as np

import rubin_scheduler.scheduler as fs
import rubin_scheduler.scheduler.Training as optional
from rubin_scheduler.model_observatory import ModelObservatory


class BlackTraining:
    def __init__(self, preferences=[1], gray_train=False, custom_period=1):
        self.pref = preferences

        self.survey_length = 0.2  # days
        self.surveys = []
        survey_filters = ["r"]
        for f in survey_filters:
            self.bfs = []
            self.bfs.append(fs.Slewtime_basis_function_cost(filtername=f))
            self.bfs.append(fs.Visit_repeat_basis_function_cost(filtername=f, survey_filters=survey_filters))
            self.bfs.append(fs.Target_map_basis_function_cost(filtername=f, survey_filters=survey_filters))
            self.bfs.append(fs.Normalized_alt_basis_function_cost(filtername=f))
            self.bfs.append(fs.Hour_angle_basis_function_cost())
            self.bfs.append(fs.Depth_percentile_basis_function_cost())
            weights = np.array([5, 2, 1, 1, 2, 1])
            self.surveys.append(
                fs.Simple_greedy_survey_fields_cost(self.bfs, weights, filtername=f, block_size=10)
            )

    def de_opt(
        self,
        n_p,
        F,
        cr,
        max_iter,
        D,
        domain,
        load_candidate_solution,
        gray_trianing=False,
    ):
        self.D = D
        self.domain = domain
        self.optimizer = optional.DE_optimizer(
            self,
            n_p,
            F,
            cr,
            max_iter,
            gray_training=gray_trianing,
            load_candidate_solution=load_candidate_solution,
        )

    def target(self, x):
        x[0] = 5  # reduce redundant solutions
        for survey in self.surveys:
            survey.basis_weights = x
        scheduler = fs.Core_scheduler_cost(self.surveys)
        observatory = ModelObservatory()
        observatory, scheduler, observations = fs.sim_runner(
            observatory, scheduler, survey_length=self.survey_length
        )
        return -1 * fs.simple_performance_measure(observations, self.pref)

    def refined_individual(self):
        return np.zeros(self.D)


n_p = 50  # number of candidate solutions that are supposed to explore the space of solution in each iteration, rule of thumb: ~10*D
F = 0.8  # algorithm meta parameter (mutation factor that determines the amount of change for the derivation of candidate solutions of the next iteration)
cr = 0.8  # algorithm meta parameter (crossover rate that determines the rate of mixing of previous candidates to make new candidates)
max_iter = 100  # maximum number of iterations. maximum number of function evaluations = n_p * max_iter,
domain = np.array(
    [[0, 10], [0, 10], [0, 10], [0, 10], [0, 10], [0, 10]]
)  # Final solution would lie in this domain
D = 6  # weights dimension


train = BlackTraining()
train.de_opt(n_p, F, cr, max_iter, D, domain, load_candidate_solution=False)



__all__ = ("dark_sky",)

import os

import healpy as hp
import numpy as np

from rubin_scheduler.data import get_data_dir


def dark_sky(nside=32):
    """Load an array of HEALpix maps that have the darkest expected sky
    backgrounds per filter.

    Parameters
    ----------
    nside : `int` (32)
        Desired nside resolution (default=32).

    Returns
    -------
    dark_sky_data : `np.ndarray`
        Named array with dark sky data for each band.
    """
    if not hasattr(dark_sky, "data"):
        # Load up the data
        data_dir = get_data_dir()
        data = np.load(os.path.join(data_dir, "skybrightness_pre", "dark_maps.npz"))
        dark_sky.data = data["dark_maps"].copy()
        data.close()

    dark_sky_data = np.empty(hp.nside2npix(nside), dtype=dark_sky.data.dtype)

    for band in dark_sky_data.dtype.names:
        dark_sky_data[band] = hp.pixelfunc.ud_grade(dark_sky.data[band], nside_out=nside)

    return dark_sky_data



import unittest

import numpy as np

import rubin_scheduler.utils as utils
from rubin_scheduler.scheduler.model_observatory import ModelObservatory


class KindaClouds:
    """Dummy class that always sets the clouds level to a unique float"""

    def __call__(self, mjd):
        return 0.23659


class ArbSeeing:
    """Dummy class to always return a specific seeing value"""

    def __call__(self, mjd):
        fwhm_500 = 1.756978
        return fwhm_500


class TestModelObservatory(unittest.TestCase):
    def test_replace(self):
        """test that we can replace default downtimes, seeing, and clouds"""

        mjd_start = utils.survey_start_mjd()
        mo_default = ModelObservatory(mjd_start=mjd_start)
        # Never load too many nights of sky
        mo_default.sky_model.load_length = 10.0
        cond_default = mo_default.return_conditions()

        # Define new downtimes
        downtimes = np.zeros(2, dtype=list(zip(["start", "end"], [float, float])))
        downtimes["start"] = np.array([1, 10]) + mjd_start
        downtimes["end"] = np.array([2, 11]) + mjd_start

        seeing_data = ArbSeeing()
        cloud_data = KindaClouds()

        mo_new = ModelObservatory(
            mjd_start=mjd_start,
            seeing_data=seeing_data,
            cloud_data=cloud_data,
            downtimes=downtimes,
        )
        # Never load too many nights of sky
        mo_new.sky_model.load_length = 10.0
        cond_new = mo_new.return_conditions()

        # Make sure the internal downtimes are different
        assert ~np.array_equal(mo_default.downtimes, mo_new.downtimes)

        # Make sure seeing is not the same
        diff = cond_default.fwhm_eff["r"] - cond_new.fwhm_eff["r"]
        assert np.nanmin(np.abs(diff)) > 0

        # Make sure cloudyness level is not the same
        assert cond_default.bulk_cloud != cond_new.bulk_cloud


if __name__ == "__main__":
    unittest.main()



__author__ = "Elahe"

import numpy as np


class DeOptimizer:
    def __init__(
        self,
        evaluator,
        population_size,
        f,
        cr,
        max_iter,
        strategy=9,
        vtr=-1e99,
        eps=0,
        show_progress=1,
        monitor_cycle=np.inf,
        gray_training=False,
        load_candidate_solution=False,
    ):
        self.show_progress = show_progress
        self.evaluator = evaluator
        self.population_size = population_size
        self.f = f
        self.cr = cr
        self.max_iter = max_iter
        self.strategy = strategy

        if strategy > 5:
            self.st = strategy - 5
        else:
            self.st = strategy

        self.monitor_cycle = monitor_cycle
        self.D = evaluator.D
        self.eps = eps
        self.vtr = vtr

        # if population slightly changes while optimization or not
        self.gray_training = gray_training
        self.load_candidate_solution = load_candidate_solution

        self.optimize()

    def optimize(self):
        # initialize optimization
        self.initialize_optimization()
        # initialize the population
        if self.load_candidate_solution:
            self.load_generation()
        if not self.load_candidate_solution:
            self.make_random_population()

        # score initial population
        # self.score_population()
        # update progress parameters
        self.init_progress()

        # iterations
        while not self.terminate():
            self.evolve()
            self.update_progress()
            self.print_status()
            self.monitor()
            if self.gray_training:
                self.swap_population()
                self.update_progress()

    def initialize_optimization(self):
        # Initial pop features
        self.population = np.zeros((self.population_size, self.D))
        self.scores = 1e99 * np.ones(self.population_size)
        self.count = 0
        self.nfeval = 0
        self.after_score_pop = np.zeros((self.population_size, self.D))

    def make_random_population(self):
        for indiv in range(self.population_size):
            self.population[indiv, :] = self.make_random_individual()

    def make_random_individual(self):
        delta = self.evaluator.domain[:, 1] - self.evaluator.domain[:, 0]
        offset = self.evaluator.domain[:, 0]
        ind = np.multiply(np.random.rand(1, self.D), delta) + offset
        return ind

    def score_population(self):
        for indiv in range(self.population_size):
            self.scores[indiv] = self.evaluator.target(self.population[indiv, :])
            self.nfeval += 1
            """change individuals for the refined values by gray training """
            self.after_score_pop[indiv, :] = self.evaluator.refined_individual()
            self.print_ind(
                indiv,
                self.scores[indiv],
                self.population[indiv, :],
                self.after_score_pop[indiv, :],
            )

    def init_progress(self):
        self.best_index = np.argmin(self.scores)
        self.best_val = self.scores[self.best_index]
        self.best_ind = self.population[self.best_index, :]
        self.mean_val = np.mean(self.scores)

    def update_progress(self):
        # bests
        temp_best_index = np.argmin(self.scores)
        temp_best_score = self.scores[temp_best_index]
        # preserving best ever individual
        if temp_best_score < self.best_val:
            print(temp_best_score, self.best_val)
            self.best_index = temp_best_index
            self.best_val = temp_best_score
            self.best_ind = self.population[self.best_index, :]
            new_mean_val = np.mean(self.scores)
            self.obj_prog = self.mean_val - new_mean_val
            self.mean_val = new_mean_val
        # algorithm parameters
        self.count += 1

    def evolve(self):
        # Candidate
        self.ui = self.cal_trials()
        self.after_score_ui = np.zeros(np.shape(self.ui))
        # Selection
        for trial_indx in range(self.population_size):
            trial_score = self.score_trial(trial_indx)

            if trial_score < self.scores[trial_indx]:
                self.population[trial_indx, :] = self.ui[trial_indx, :]
                self.scores[trial_indx] = trial_score
                self.after_score_pop[trial_indx, :] = self.after_score_ui[trial_indx, :]

            self.print_ind(
                trial_indx,
                self.scores[trial_indx],
                self.population[trial_indx, :],
                self.after_score_pop[trial_indx, :],
            )
        self.save_last_generation()  # most recent generation

    def score_trial(self, trial_indx):
        tempval = self.evaluator.target(self.ui[trial_indx, :])
        self.nfeval += 1
        if self.gray_training:
            self.after_score_ui[trial_indx, :] = self.evaluator.refined_individual()
        return tempval

    def monitor(self):
        if self.count != 0 and self.count % self.monitor_cycle == 0:
            self.monitor_score = self.best_val
            self.monitor_indiv = self.best_ind
            self.monitor_mean = self.mean_val

    def swap_population(self):
        # regularize refined population before swap (gene-wise swap)
        for gene in range(self.D):
            self_val = self.after_score_pop[:, gene]
            lower_alt_val = self.evaluator.domain[gene, 0]
            upper_alt_val = self.evaluator.domain[gene, 1]
            lower_bound = self_val >= lower_alt_val
            upper_bound = self_val <= upper_alt_val
            self_val = np.where(lower_bound, self_val, lower_alt_val)
            self_val = np.where(upper_bound, self_val, upper_alt_val)
            # swap individuals:
            self.population[:, gene] = self_val
        # modify scores
        self.score_population()

    def cal_trials(self):
        popold = self.population

        rot = np.arange(0, self.population_size)
        rotd = np.arange(0, self.D)
        ind = np.random.permutation(4)

        a1 = np.random.permutation(self.population_size)
        rt = (rot + ind[0]) % self.population_size
        a2 = a1[rt]
        rt = (rot + ind[1]) % self.population_size
        a3 = a2[rt]
        rt = (rot + ind[2]) % self.population_size
        a4 = a3[rt]
        rt = (rot + ind[3]) % self.population_size
        a5 = a4[rt]

        pm1 = self.population[a1, :]
        pm2 = self.population[a2, :]
        pm3 = self.population[a3, :]
        pm4 = self.population[a4, :]
        pm5 = self.population[a5, :]

        pop_of_best_ind = np.zeros((self.population_size, self.D))
        for i in range(0, self.population_size):
            pop_of_best_ind[i] = self.best_ind

        cr_decision = np.random.rand(self.population_size, self.D) < self.cr

        if self.strategy > 5:
            cr_decision = np.sort(np.transpose(cr_decision))
            for i in range(0, self.population_size):
                n = np.floor(np.random.rand(1) * self.D)
                if n > 0:
                    rtd = (rotd + n) % self.D
                    rtd = rtd.astype(int)
                    cr_decision[:, i] = cr_decision[rtd, i]
            cr_decision = np.transpose(cr_decision)
        mpo = cr_decision < 0.5
        ui = 0
        if self.st == 1:
            dif = self.f * (pm1 - pm2)
            ui = pop_of_best_ind + dif
            ui = self.regularize_candidate(ui, pop_of_best_ind)
            ui = np.multiply(popold, mpo) + np.multiply(ui, cr_decision)
        elif self.st == 2:
            dif = self.f * (pm1 - pm2)
            ui = pm3 + dif
            ui = self.regularize_candidate(ui, pop_of_best_ind)
            ui = np.multiply(popold, mpo) + np.multiply(ui, cr_decision)
        elif self.st == 3:
            dif = self.f * (pop_of_best_ind - popold + pm1 - pm2)
            ui = popold + dif
            ui = self.regularize_candidate(ui, pop_of_best_ind)
            ui = np.multiply(popold, mpo) + np.multiply(ui, cr_decision)
        elif self.st == 4:
            dif = self.f * (pm1 - pm2 + pm3 - pm4)
            ui = pop_of_best_ind + dif
            ui = self.regularize_candidate(ui, pop_of_best_ind)
            ui = np.multiply(popold, mpo) + np.multiply(ui, cr_decision)
        elif self.st == 5:
            dif = self.f * (pm1 - pm2 + pm3 - pm4)
            ui = pm5 + dif
            ui = self.regularize_candidate(ui, pop_of_best_ind)
            ui = np.multiply(popold, mpo) + np.multiply(ui, cr_decision)
        return ui

    def regularize_candidate(self, candidate, alternative):
        lower_bound = self.evaluator.domain[:, 0]
        upper_bound = self.evaluator.domain[:, 1]
        candidate = np.where(candidate >= lower_bound, candidate, alternative)
        regularized_cand = np.where(candidate <= upper_bound, candidate, alternative)
        return regularized_cand

    def terminate(self):
        # Termination 1 : By maxiter
        if self.count >= self.max_iter:
            self.termination = 1
            return True
        """
    #Termination 2 : By Value to reach
        if self.best_val < self.vtr :
            self.termination = 2
            return True
    #Termination 3 : By monitor cycle change in the objective
        if  self.monitor_obj_change < self.eps:
            self.termination = 3
            return True
            """
        return False

    def save_last_generation(self):
        np.save("last_gen_pop", self.population)
        np.save("last_gen_scr", self.scores)

    def load_generation(self):
        try:
            temp_population = np.load("last_gen_pop.npy")
            if np.shape(temp_population)[0] != self.population_size:
                print(
                    "Previous generation is not of the same size of new setting, DE starts with a random initialization"
                )
                self.load_candidate_solution = False
            elif np.shape(temp_population)[1] != self.D:
                print(
                    "Previous solution is not of the same size of new solution, DE starts with a random initialization"
                )
                self.load_candidate_solution = False
            else:
                self.population = temp_population
                self.scores = np.load("last_gen_scr.npy")
                print("Warm start: DE starts with a previously evolved population")

        except:
            print("No previous generation is available, DE starts with a random initialization")
            self.load_candidate_solution = False

    def print_ind(self, ind_index, score, indiv, refined_indiv):
        # print("{}: Objective: {},\nCandidate: {},\nRefined candidate: {}".format(ind_index +1, score, indiv, refined_indiv))
        print("{}: Objective: {},\nCandidate: {}".format(ind_index + 1, -1.0 * score, indiv))
        with open("Output/Output.txt", "a") as text_file:
            text_file.write("{}: Objective: {},\nCandidate: {}\n".format(ind_index + 1, -1.0 * score, indiv))
            text_file.close()

    def print_status(self):
        print("********************************************************************************************")
        print(
            "iter {}:best performance: {}, best candidate no.: {}\nbest candidate: {}".format(
                self.count, self.best_val * -1, self.best_index + 1, self.best_ind
            )
        )
        print("")
        with open("Output/Output.txt", "a") as text_file:
            text_file.write(
                "********************************************************************************************"
            )
            text_file.write(
                "iter {}:best performance: {}, best candidate no.: {}\nbest candidate: {}\n".format(
                    self.count, self.best_val * -1, self.best_index + 1, self.best_ind
                )
            )
            text_file.close()

    def final_print(self):
        print("")
        print("* Problem Specifications")
        print("Date           : {}".format(self.evaluator.scheduler.Date))
        print("Preferences    : {}".format(self.evaluator.pref))
        print("")
        print("* DE parameters")
        print("F          : {}".format(self.f))
        print("Cr         : {}".format(self.cr))
        print("Pop Size   : {}".format(self.population_size))
        print("No. of eval: {}".format(self.nfeval))
        print("No. of Iter: {}".format(self.count))
        print("Termination: {}".format(self.termination))
        print("eps        : {}".format(self.eps))
        print("vtr        : {}".format(self.vtr))
        print("")
        with open("Output/Output.txt", "a") as text_file:
            text_file.write("\n* Problem Specifications\n")
            text_file.close()
            text_file.write("Date           : {}\n".format(self.evaluator.scheduler.Date))
            text_file.close()
            text_file.write("Preferences    : {}\n".format(self.evaluator.pref))
            text_file.close()
            text_file.write("\n\n")
            text_file.close()
            text_file.write("* DE parameters\n")
            text_file.write("F          : {}\n".format(self.f))
            text_file.close()
            text_file.write("Cr         : {}\n".format(self.cr))
            text_file.close()
            text_file.write("Pop Size   : {}\n".format(self.population_size))
            text_file.close()
            text_file.write("No. of eval: {}\n".format(self.nfeval))
            text_file.close()
            text_file.write("No. of Iter: {}\n".format(self.count))
            text_file.close()
            text_file.write("Termination: {}\n".format(self.termination))
            text_file.close()
            text_file.write("eps        : {}\n".format(self.eps))
            text_file.close()
            text_file.write("vtr        : {}\n".format(self.vtr))
            text_file.close()



__all__ = ("ScriptedSurvey", "PairsSurveyScripted")

import logging

import numpy as np

import rubin_scheduler.scheduler.features as features
from rubin_scheduler.scheduler.surveys import BaseSurvey
from rubin_scheduler.scheduler.utils import empty_observation, set_default_nside
from rubin_scheduler.utils import _approx_ra_dec2_alt_az, ra_dec2_hpid

log = logging.getLogger(__name__)


class ScriptedSurvey(BaseSurvey):
    """
    Take a set of scheduled observations and serve them up.

    Parameters
    ----------
    id_start : `int` (1)
        The integer to start the "scripted id" field with. Bad things could happen
        if you have multiple scripted survey objects with the same scripted IDs.
    return_n_limit : `int` (10)
        The maximum number of observations to return. Set to high and your block
        of scheduled observations can run into twilight time.
    """

    def __init__(
        self,
        basis_functions,
        reward=1e6,
        ignore_obs="dummy",
        nside=None,
        detailers=None,
        id_start=1,
        return_n_limit=10,
        survey_name=None,
    ):
        """"""
        if nside is None:
            nside = set_default_nside()

        self.extra_features = {}
        self.nside = nside
        self.reward_val = reward
        self.reward = -np.inf
        self.id_start = id_start
        self.return_n_limit = return_n_limit
        super(ScriptedSurvey, self).__init__(
            basis_functions=basis_functions,
            ignore_obs=ignore_obs,
            nside=nside,
            detailers=detailers,
            survey_name=survey_name,
        )
        self.clear_script()

    def add_observations_array(self, observations_array_in, observations_hpid_in):
        if self.obs_wanted is not None:
            # toss out things that should be ignored
            to_ignore = np.in1d(observations_array_in["note"], self.ignore_obs)
            observations_array = observations_array_in[~to_ignore]

            good = np.in1d(observations_hpid_in["ID"], observations_array["ID"])
            observations_hpid = observations_hpid_in[good]

            for feature in self.extra_features:
                self.extra_features[feature].add_observations_array(observations_array, observations_hpid)
            for bf in self.extra_basis_functions:
                self.extra_basis_functions[bf].add_observations_array(observations_array, observations_hpid)
            for bf in self.basis_functions:
                bf.add_observations_array(observations_array, observations_hpid)
            for detailer in self.detailers:
                detailer.add_observations_array(observations_array, observations_hpid)

            # If scripted_id, note, and filter match, then consider the observation completed.
            completed = np.char.add(
                observations_array["scripted_id"].astype(str),
                observations_array["note"],
            )
            completed = np.char.add(completed, observations_array["filter"])

            wanted = np.char.add(self.obs_wanted["scripted_id"].astype(str), self.obs_wanted["note"])
            wanted = np.char.add(wanted, self.obs_wanted["filter"])

            indx = np.in1d(wanted, completed)
            self.obs_wanted["observed"][indx] = True
            self.scheduled_obs = self.obs_wanted["mjd"][~self.obs_wanted["observed"]]

    def add_observation(self, observation, indx=None, **kwargs):
        """Check if observation matches a scripted observation"""
        if self.obs_wanted is not None:
            # From base class
            checks = [io not in str(observation["note"]) for io in self.ignore_obs]
            if all(checks):
                for feature in self.extra_features:
                    self.extra_features[feature].add_observation(observation, **kwargs)
                for bf in self.basis_functions:
                    bf.add_observation(observation, **kwargs)
                for detailer in self.detailers:
                    detailer.add_observation(observation, **kwargs)
                self.reward_checked = False

                # find the index
                indx = np.searchsorted(self.obs_wanted["scripted_id"], observation["scripted_id"])
                # If it matches scripted_id, note, and filter, mark it as observed and update scheduled observation list.
                if indx < self.obs_wanted["scripted_id"].size:
                    if (
                        (self.obs_wanted["scripted_id"][indx] == observation["scripted_id"])
                        & (self.obs_wanted["note"][indx] == observation["note"])
                        & (self.obs_wanted["filter"][indx] == observation["filter"])
                    ):
                        self.obs_wanted["observed"][indx] = True
                        self.scheduled_obs = self.obs_wanted["mjd"][~self.obs_wanted["observed"]]

    def calc_reward_function(self, conditions):
        """If there is an observation ready to go, execute it, otherwise, -inf"""
        observation = self._check_list(conditions)
        if observation is None:
            self.reward = -np.inf
        else:
            self.reward = self.reward_val
        return self.reward

    def _slice2obs(self, obs_row):
        """take a slice and return a full observation object"""
        observation = empty_observation()
        for key in [
            "RA",
            "dec",
            "filter",
            "exptime",
            "nexp",
            "note",
            "target",
            "rotSkyPos",
            "rotTelPos",
            "flush_by_mjd",
            "scripted_id",
        ]:
            observation[key] = obs_row[key]
        return observation

    def _check_alts_ha(self, observation, conditions):
        """Given scheduled observations, check which ones can be done in current conditions.

        Parameters
        ----------
        observation : np.array
            An array of scheduled observations. Probably generated with rubin_scheduler.scheduler.utils.scheduled_observation
        """
        # Just do a fast ra,dec to alt,az conversion.
        alt, az = _approx_ra_dec2_alt_az(
            observation["RA"],
            observation["dec"],
            conditions.site.latitude_rad,
            None,
            conditions.mjd,
            lmst=conditions.lmst,
        )
        HA = conditions.lmst - observation["RA"] * 12.0 / np.pi
        HA[np.where(HA > 24)] -= 24
        HA[np.where(HA < 0)] += 24
        in_range = np.where(
            (alt < observation["alt_max"])
            & (alt > observation["alt_min"])
            & ((HA > observation["HA_max"]) | (HA < observation["HA_min"]))
            & (conditions.sun_alt < observation["sun_alt_max"])
        )[0]
        return in_range

    def _check_list(self, conditions):
        """Check to see if the current mjd is good"""
        observations = None
        if self.obs_wanted is not None:
            # Scheduled observations that are in the right time window and have not been executed
            in_time_window = np.where(
                (self.mjd_start < conditions.mjd)
                & (self.obs_wanted["flush_by_mjd"] > conditions.mjd)
                & (~self.obs_wanted["observed"])
            )[0]

            if np.size(in_time_window) > 0:
                pass_checks = self._check_alts_ha(self.obs_wanted[in_time_window], conditions)
                matches = in_time_window[pass_checks]

                # Also check that the filters are mounted
                match2 = np.isin(self.obs_wanted["filter"][matches], conditions.mounted_filters)
                matches = matches[match2]

            else:
                matches = []

            if np.size(matches) > 0:
                # Do not return too many observations
                if np.size(matches) > self.return_n_limit:
                    matches = matches[0 : self.return_n_limit]
                observations = self.obs_wanted[matches]

        return observations

    def clear_script(self):
        """set an empty list to serve up"""
        self.obs_wanted = None
        self.mjd_start = None
        self.scheduled_obs = None

    def set_script(self, obs_wanted):
        """
        Parameters
        ----------
        obs_wanted : np.array
            The observations that should be executed. Needs to have columns with dtype names:
            Should be from lsst.sim.scheduler.utils.scheduled_observation
        mjds : np.array
            The MJDs for the observaitons, should be same length as obs_list
        mjd_tol : float (15.)
            The tolerance to consider an observation as still good to observe (min)
        """

        self.obs_wanted = obs_wanted

        self.obs_wanted.sort(order=["mjd", "filter"])
        # Give each desired observation a unique "scripted ID". To be used for
        # matching and logging later.
        self.obs_wanted["scripted_id"] = np.arange(self.id_start, self.id_start + np.size(self.obs_wanted))
        # Update so if we set the script again the IDs will not be reused.
        self.id_start = np.max(self.obs_wanted["scripted_id"]) + 1

        self.mjd_start = self.obs_wanted["mjd"] - self.obs_wanted["mjd_tol"]
        # Here is the atribute that core scheduler checks to broadcast scheduled observations
        # in the conditions object.
        self.scheduled_obs = self.obs_wanted["mjd"]

    def generate_observations_rough(self, conditions):
        observations = self._check_list(conditions)
        observations = [self._slice2obs(obs) for obs in observations]

        return observations


class PairsSurveyScripted(ScriptedSurvey):
    """Check if incoming observations will need a pair in 30 minutes. If so, add to the queue"""

    def __init__(
        self,
        basis_functions,
        filt_to_pair="griz",
        dt=40.0,
        ttol=10.0,
        reward_val=101.0,
        note="scripted",
        ignore_obs="ack",
        min_alt=30.0,
        max_alt=85.0,
        lat=-30.2444,
        moon_distance=30.0,
        max_slew_to_pair=15.0,
        nside=None,
        survey_name=None,
    ):
        """
        Parameters
        ----------
        filt_to_pair : str (griz)
            Which filters to try and get pairs of
        dt : float (40.)
            The ideal gap between pairs (minutes)
        ttol : float (10.)
            The time tolerance when gathering a pair (minutes)
        """
        if nside is None:
            nside = set_default_nside()

        super(PairsSurveyScripted, self).__init__(
            basis_functions=basis_functions,
            ignore_obs=ignore_obs,
            min_alt=min_alt,
            max_alt=max_alt,
            nside=nside,
            survey_name=survey_name,
        )

        self.lat = np.radians(lat)
        self.note = note
        self.ttol = ttol / 60.0 / 24.0
        self.dt = dt / 60.0 / 24.0  # To days
        self.max_slew_to_pair = max_slew_to_pair  # in seconds
        self._moon_distance = np.radians(moon_distance)

        self.extra_features = {}
        self.extra_features["Pair_map"] = features.Pair_in_night(filtername=filt_to_pair)

        self.reward_val = reward_val
        self.filt_to_pair = filt_to_pair
        # list to hold observations
        self.observing_queue = []
        # make ignore_obs a list
        if type(self.ignore_obs) is str:
            self.ignore_obs = [self.ignore_obs]

    def add_observation(self, observation, indx=None, **kwargs):
        """Add an observed observation"""
        # self.ignore_obs not in str(observation['note'])
        to_ignore = np.any([ignore in str(observation["note"]) for ignore in self.ignore_obs])
        log.debug(
            "[Pairs.add_observation]: %s: %s: %s",
            to_ignore,
            str(observation["note"]),
            self.ignore_obs,
        )
        log.debug("[Pairs.add_observation.queue]: %s", self.observing_queue)
        if not to_ignore:
            # Update my extra features:
            for feature in self.extra_features:
                if hasattr(self.extra_features[feature], "add_observation"):
                    self.extra_features[feature].add_observation(observation, indx=indx)
            self.reward_checked = False

            # Check if this observation needs a pair
            # XXX--only supporting single pairs now. Just start up another scripted survey
            # to grab triples, etc? Or add two observations to queue at a time?
            # keys_to_copy = ['RA', 'dec', 'filter', 'exptime', 'nexp']
            if (observation["filter"][0] in self.filt_to_pair) and (
                np.max(self.extra_features["Pair_map"].feature[indx]) < 1
            ):
                obs_to_queue = empty_observation()
                for key in observation.dtype.names:
                    obs_to_queue[key] = observation[key]
                # Fill in the ideal time we would like this observed
                log.debug("Observation MJD: %.4f (dt=%.4f)", obs_to_queue["mjd"], self.dt)
                obs_to_queue["mjd"] += self.dt
                self.observing_queue.append(obs_to_queue)
        log.debug("[Pairs.add_observation.queue.size]: %i", len(self.observing_queue))
        for obs in self.observing_queue:
            log.debug("[Pairs.add_observation.queue]: %s", obs)

    def _purge_queue(self, conditions):
        """Remove any pair where it's too late to observe it"""
        # Assuming self.observing_queue is sorted by MJD.
        if len(self.observing_queue) > 0:
            stale = True
            in_window = np.abs(self.observing_queue[0]["mjd"] - conditions.mjd) < self.ttol
            log.debug("Purging queue")
            while stale:
                # If the next observation in queue is past the window, drop it
                if (self.observing_queue[0]["mjd"] < conditions.mjd) & (~in_window):
                    log.debug(
                        "Past the window: obs_mjd=%.4f (current_mjd=%.4f)",
                        self.observing_queue[0]["mjd"],
                        conditions.mjd,
                    )
                    del self.observing_queue[0]
                # If we are in the window, but masked, drop it
                elif (in_window) & (~self._check_mask(self.observing_queue[0], conditions)):
                    log.debug("Masked")
                    del self.observing_queue[0]
                # If in time window, but in alt exclusion zone
                elif (in_window) & (~self._check_alts(self.observing_queue[0], conditions)):
                    log.debug("in alt exclusion zone")
                    del self.observing_queue[0]
                else:
                    stale = False
                # If we have deleted everything, break out of where
                if len(self.observing_queue) == 0:
                    stale = False

    def _check_alts(self, observation, conditions):
        result = False
        # Just do a fast ra,dec to alt,az conversion. Can use LMST from a feature.

        alt, az = _approx_ra_dec2_alt_az(
            observation["RA"],
            observation["dec"],
            self.lat,
            None,
            conditions.mjd,
            lmst=conditions.lmst,
        )
        in_range = np.where((alt < self.max_alt) & (alt > self.min_alt))[0]
        if np.size(in_range) > 0:
            result = True
        return result

    def _check_mask(self, observation, conditions):
        """Check that the proposed observation is not currently masked for some reason on the sky map.
        True if the observation is good to observe
        False if the proposed observation is masked
        """

        hpid = np.max(ra_dec2_hpid(self.nside, observation["RA"], observation["dec"]))
        skyval = conditions.M5Depth[observation["filter"][0]][hpid]

        if skyval > 0:
            return True
        else:
            return False

    def calc_reward_function(self, conditions):
        self._purge_queue(conditions)
        result = -np.inf
        self.reward = result
        log.debug("Pair - calc_reward_func")
        for indx in range(len(self.observing_queue)):
            check = self._check_observation(self.observing_queue[indx], conditions)
            log.debug("%s: %s", check, self.observing_queue[indx])
            if check[0]:
                result = self.reward_val
                self.reward = self.reward_val
                break
            elif not check[1]:
                break

        self.reward_checked = True
        return result

    def _check_observation(self, observation, conditions):
        delta_t = observation["mjd"] - conditions.mjd
        log.debug(
            "Check_observation: obs_mjd=%.4f (current_mjd=%.4f, delta=%.4f, tol=%.4f)",
            observation["mjd"],
            conditions.mjd,
            delta_t,
            self.ttol,
        )
        obs_hp = ra_dec2_hpid(self.nside, observation["RA"], observation["dec"])
        slewtime = conditions.slewtime[obs_hp[0]]
        in_slew_window = slewtime <= self.max_slew_to_pair or delta_t < 0.0
        in_time_window = np.abs(delta_t) < self.ttol

        if conditions.current_filter is None:
            infilt = True
        else:
            infilt = conditions.current_filter in self.filt_to_pair

        is_observable = self._check_mask(observation, conditions)
        valid = in_time_window & infilt & in_slew_window & is_observable
        log.debug("Pair - observation: %s " % observation)
        log.debug(
            "Pair - check[%s]: in_time_window[%s] infilt[%s] in_slew_window[%s] is_observable[%s]"
            % (valid, in_time_window, infilt, in_slew_window, is_observable)
        )

        return (valid, in_time_window, infilt, in_slew_window, is_observable)

    def generate_observations(self, conditions):
        # Toss anything in the queue that is too old to pair up:
        self._purge_queue(conditions)
        # Check for something I want a pair of
        result = []
        # if len(self.observing_queue) > 0:
        log.debug("Pair - call")
        for indx in range(len(self.observing_queue)):
            check = self._check_observation(self.observing_queue[indx], conditions)

            if check[0]:
                result = self.observing_queue.pop(indx)
                result["note"] = "pair(%s)" % self.note
                # Make sure we don't change filter if we don't have to.
                if conditions.current_filter is not None:
                    result["filter"] = conditions.current_filter
                # Make sure it is observable!
                # if self._check_mask(result):
                result = [result]
                break
            elif not check[1]:
                # If this is not in time window and queue is chronological, none will be...
                break

        return result



import unittest

from rubin_scheduler.site_models import Almanac


class TestAlmanac(unittest.TestCase):
    def test_alm(self):
        alma = Almanac()

        mjd = 59853.35

        # Dead simple make sure the things load.
        planets = alma.get_planet_positions(mjd)
        sun = alma.get_sunset_info(mjd)
        moon = alma.get_sun_moon_positions(mjd)
        indx = alma.mjd_indx(mjd)


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "BaseDetailer",
    "ZeroRotDetailer",
    "Comcam90rotDetailer",
    "Rottep2RotspDesiredDetailer",
    "CloseAltDetailer",
    "TakeAsPairsDetailer",
    "TwilightTripleDetailer",
    "SpiderRotDetailer",
    "FlushForSchedDetailer",
    "FilterNexp",
    "FixedSkyAngleDetailer",
)

import copy

import numpy as np

from rubin_scheduler.scheduler.utils import IntRounded
from rubin_scheduler.utils import _angular_separation, _approx_altaz2pa, _approx_ra_dec2_alt_az


class BaseDetailer:
    """
    A Detailer is an object that takes a list of proposed observations and adds "details" to them. The
    primary purpose is that the Markov Decision Process does an excelent job selecting RA,Dec,filter
    combinations, but we may want to add additional logic such as what to set the camera rotation angle
    to, or what to use for an exposure time. We could also modify the order of the proposed observations.
    For Deep Drilling Fields, a detailer could be useful for computing dither positions and modifying
    the exact RA,Dec positions.
    """

    def __init__(self, nside=32):
        """"""
        # Dict to hold all the features we want to track
        self.survey_features = {}
        self.nside = nside

    def add_observations_array(self, observations_array, observations_hpid):
        """Like add_observation, but for loading a whole array of observations at a time"""

        for feature in self.survey_features:
            self.survey_features[feature].add_observations_array(observations_array, observations_hpid)

    def add_observation(self, observation, indx=None):
        """
        Parameters
        ----------
        observation : `np.array`
            An array with information about the input observation
        indx : `np.array`
            The indices of the healpix map that the observation overlaps with
        """
        for feature in self.survey_features:
            self.survey_features[feature].add_observation(observation, indx=indx)

    def __call__(self, observation_list, conditions):
        """
        Parameters
        ----------
        observation_list : `list` of observations
            The observations to detail.
        conditions : `rubin_scheduler.scheduler.conditions` object

        Returns
        -------
        List of observations.
        """

        return observation_list


class Rottep2RotspDesiredDetailer(BaseDetailer):
    """Convert all the rotTelPos values to rotSkyPos_desired"""

    def __call__(self, observation_list, conditions):
        obs_array = np.concatenate(observation_list)

        alt, az = _approx_ra_dec2_alt_az(
            obs_array["RA"],
            obs_array["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)

        rot_sky_pos_desired = (obs_array["rotTelPos"] - obs_pa) % (2.0 * np.pi)

        for obs, rotsp_d in zip(observation_list, rot_sky_pos_desired):
            obs["rotTelPos_backup"] = obs["rotTelPos"] + 0
            obs["rotTelPos"] = np.nan
            obs["rotSkyPos"] = np.nan
            obs["rotSkyPos_desired"] = rotsp_d

        return observation_list


class ZeroRotDetailer(BaseDetailer):
    """
    Detailer to set the camera rotation to be apporximately zero in rotTelPos.
    Because it can never be written too many times:
    rotSkyPos = rotTelPos - ParallacticAngle
    But, wait, what? Is it really the other way?
    """

    def __call__(self, observation_list, conditions):
        # XXX--should I convert the list into an array and get rid of this loop?
        for obs in observation_list:
            alt, az = _approx_ra_dec2_alt_az(
                obs["RA"],
                obs["dec"],
                conditions.site.latitude_rad,
                conditions.site.longitude_rad,
                conditions.mjd,
            )
            obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)
            obs["rotSkyPos"] = obs_pa

        return observation_list


class SpiderRotDetailer(BaseDetailer):
    """
    Set the camera rotation to +/- 45 degrees so diffraction spikes align along chip rows
    and columns
    """

    def __call__(self, observation_list, conditions):
        indx = int(conditions.night % 2)
        rot_tel_pos = np.radians([45.0, 315.0][indx])

        for obs in observation_list:
            obs["rotSkyPos"] = np.nan
            obs["rot_tel_pos"] = rot_tel_pos

        return observation_list


class Comcam90rotDetailer(BaseDetailer):
    """
    Detailer to set the camera rotation so rotSkyPos is 0, 90, 180, or 270 degrees. Whatever
    is closest to rotTelPos of zero.
    """

    def __call__(self, observation_list, conditions):
        favored_rot_sky_pos = np.radians([0.0, 90.0, 180.0, 270.0, 360.0]).reshape(5, 1)
        obs_array = np.concatenate(observation_list)
        alt, az = _approx_ra_dec2_alt_az(
            obs_array["RA"],
            obs_array["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        parallactic_angle = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)
        # If we set rotSkyPos to parallactic angle, rotTelPos will be zero. So, find the
        # favored rotSkyPos that is closest to PA to keep rotTelPos as close as possible to zero.
        ang_diff = np.abs(parallactic_angle - favored_rot_sky_pos)
        min_indxs = np.argmin(ang_diff, axis=0)
        # can swap 360 and zero if needed?
        final_rot_sky_pos = favored_rot_sky_pos[min_indxs]
        # Set all the observations to the proper rotSkyPos
        for rsp, obs in zip(final_rot_sky_pos, observation_list):
            obs["rotSkyPos"] = rsp

        return observation_list


class FixedSkyAngleDetailer(BaseDetailer):
    """Detailer to force a specific sky angle.

    Parameters
    ----------
    sky_angle : `float`, optional
        Desired sky angle (default = 0, in degrees).
    """

    def __init__(self, sky_angle=0.0, nside=32):
        super().__init__(nside=nside)

        self.sky_angle = np.radians(sky_angle)

    def __call__(self, observation_list, conditions):
        for observation in observation_list:
            observation["rotSkyPos"] = self.sky_angle

        return observation_list


class CloseAltDetailer(BaseDetailer):
    """
    re-order a list of observations so that the closest in altitude to the current pointing is first.

    Parameters
    ----------
    alt_band : `float` (10)
        The altitude band to try and stay in (degrees)
    """

    def __init__(self, alt_band=10.0):
        super(CloseAltDetailer, self).__init__()
        self.alt_band = IntRounded(np.radians(alt_band))

    def __call__(self, observation_list, conditions):
        obs_array = np.concatenate(observation_list)
        alt, az = _approx_ra_dec2_alt_az(
            obs_array["RA"],
            obs_array["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        alt_diff = np.abs(alt - conditions.tel_alt)
        in_band = np.where(IntRounded(alt_diff) <= self.alt_band)[0]
        if in_band.size == 0:
            in_band = np.arange(alt.size)

        # Find the closest in angular distance of the points that are in band
        ang_dist = _angular_separation(az[in_band], alt[in_band], conditions.tel_az, conditions.tel_alt)
        if np.size(ang_dist) == 1:
            good = 0
        else:
            good = np.min(np.where(ang_dist == ang_dist.min())[0])
        indx = in_band[good]
        result = observation_list[indx:] + observation_list[:indx]
        return result


class FlushForSchedDetailer(BaseDetailer):
    """Update the flush-by MJD to be before any scheduled observations

    Parameters
    ----------
    tol : `float`
         How much before to flush (minutes)
    """

    def __init__(self, tol=2.5):
        super(FlushForSchedDetailer, self).__init__()
        self.tol = tol / 24.0 / 60.0  # To days

    def __call__(self, observation_list, conditions):
        if np.size(conditions.scheduled_observations) > 0:
            new_flush = np.min(conditions.scheduled_observations) - self.tol
            for obs in observation_list:
                if obs["flush_by_mjd"] > new_flush:
                    obs["flush_by_mjd"] = new_flush
        return observation_list


class FilterNexp(BaseDetailer):
    """Demand one filter always be taken as a certain number of exposures"""

    def __init__(self, filtername="u", nexp=1, exptime=None):
        super(FilterNexp, self).__init__()
        self.filtername = filtername
        self.nexp = nexp
        self.exptime = exptime

    def __call__(self, observation_list, conditions):
        for obs in observation_list:
            if obs["filter"] == self.filtername:
                obs["nexp"] = self.nexp
                if self.exptime is not None:
                    obs["exptime"] = self.exptime
        return observation_list


class TakeAsPairsDetailer(BaseDetailer):
    def __init__(self, filtername="r", exptime=None, nexp_dict=None):
        """"""
        super(TakeAsPairsDetailer, self).__init__()
        self.filtername = filtername
        self.exptime = exptime
        self.nexp_dict = nexp_dict

    def __call__(self, observation_list, conditions):
        paired = copy.deepcopy(observation_list)
        if self.exptime is not None:
            for obs in paired:
                obs["exptime"] = self.exptime
        for obs in paired:
            obs["filter"] = self.filtername
            if self.nexp_dict is not None:
                obs["nexp"] = self.nexp_dict[self.filtername]
        if conditions.current_filter == self.filtername:
            for obs in paired:
                obs["note"] = obs["note"][0] + ", a"
            for obs in observation_list:
                obs["note"] = obs["note"][0] + ", b"
            result = paired + observation_list
        else:
            for obs in paired:
                obs["note"] = obs["note"][0] + ", b"
            for obs in observation_list:
                obs["note"] = obs["note"][0] + ", a"
            result = observation_list + paired
        # XXX--maybe a temp debugging thing, label what part of sequence each observation is.
        for i, obs in enumerate(result):
            obs["survey_id"] = i
        return result


class TwilightTripleDetailer(BaseDetailer):
    def __init__(self, slew_estimate=5.0, n_repeat=3, update_note=True):
        super(TwilightTripleDetailer, self).__init__()
        self.slew_estimate = slew_estimate
        self.n_repeat = n_repeat
        self.update_note = update_note

    def __call__(self, observation_list, conditions):
        obs_array = np.concatenate(observation_list)

        # Estimate how much time is left in the twilgiht block
        potential_times = np.array(
            [
                conditions.sun_n18_setting - conditions.mjd,
                conditions.sun_n12_rising - conditions.mjd,
            ]
        )

        potential_times = np.min(potential_times[np.where(potential_times > 0)]) * 24.0 * 3600.0

        # How long will observations take?
        cumulative_slew = np.arange(obs_array.size) * self.slew_estimate
        cumulative_expt = np.cumsum(obs_array["exptime"])
        cumulative_time = cumulative_slew + cumulative_expt
        # If we are way over, truncate the list before doing the triple
        if np.max(cumulative_time) > potential_times:
            max_indx = np.where(cumulative_time / self.n_repeat <= potential_times)[0]
            if np.size(max_indx) == 0:
                # Very bad magic number fudge
                max_indx = 3
            else:
                max_indx = np.max(max_indx)
                if max_indx == 0:
                    max_indx += 1
            observation_list = observation_list[0:max_indx]

        # Repeat the observations n times
        out_obs = []
        for i in range(self.n_repeat):
            sub_list = copy.deepcopy(observation_list)
            if self.update_note:
                for obs in sub_list:
                    obs["note"][0] += ", %i" % i
            out_obs.extend(sub_list)
        return out_obs



__all__ = ("Almanac",)

import os

import numpy as np
from scipy.interpolate import interp1d

from rubin_scheduler.data import get_data_dir


class Almanac:
    """Class to load and return pre-computed information about the LSST site."""

    def __init__(self, mjd_start=None, kind="quadratic"):
        # Load up the sunrise/sunset times
        data_dir = os.path.join(get_data_dir(), "site_models")

        temp = np.load(os.path.join(data_dir, "sunsets.npz"))
        self.sunsets = temp["almanac"].copy()
        temp.close()

        if mjd_start is not None:
            loc = np.searchsorted(self.sunsets["sunset"], mjd_start)
            # Set the start MJD to be night 1.
            self.sunsets["night"] -= self.sunsets["night"][loc - 1]

        # Load the sun and moon positions
        temp = np.load(os.path.join(data_dir, "sun_moon.npz"))
        self.sun_moon = temp["sun_moon_info"].copy()
        temp.close()

        self.interpolators = {
            "sun_alt": interp1d(self.sun_moon["mjd"], self.sun_moon["sun_alt"], kind=kind),
            "sun_az_x": interp1d(self.sun_moon["mjd"], np.cos(self.sun_moon["sun_az"]), kind=kind),
            "sun_az_y": interp1d(self.sun_moon["mjd"], np.sin(self.sun_moon["sun_az"]), kind=kind),
            "sun_RA_x": interp1d(self.sun_moon["mjd"], np.cos(self.sun_moon["sun_RA"]), kind=kind),
            "sun_RA_y": interp1d(self.sun_moon["mjd"], np.sin(self.sun_moon["sun_RA"]), kind=kind),
            "sun_dec": interp1d(self.sun_moon["mjd"], self.sun_moon["sun_dec"], kind=kind),
            "moon_alt": interp1d(self.sun_moon["mjd"], self.sun_moon["moon_alt"], kind=kind),
            "moon_az_x": interp1d(self.sun_moon["mjd"], np.cos(self.sun_moon["moon_az"]), kind=kind),
            "moon_az_y": interp1d(self.sun_moon["mjd"], np.sin(self.sun_moon["moon_az"]), kind=kind),
            "moon_RA_x": interp1d(self.sun_moon["mjd"], np.cos(self.sun_moon["moon_RA"]), kind=kind),
            "moon_RA_y": interp1d(self.sun_moon["mjd"], np.sin(self.sun_moon["moon_RA"]), kind=kind),
            "moon_dec": interp1d(self.sun_moon["mjd"], self.sun_moon["moon_dec"], kind=kind),
            "moon_phase": interp1d(self.sun_moon["mjd"], self.sun_moon["moon_phase"], kind=kind),
        }

        temp = np.load(os.path.join(data_dir, "planet_locations.npz"))
        self.planet_loc = temp["planet_loc"].copy()
        temp.close()

        self.planet_names = ["venus", "mars", "jupiter", "saturn"]
        self.planet_interpolators = {}
        for pn in self.planet_names:
            self.planet_interpolators[pn + "_RA_x"] = interp1d(
                self.planet_loc["mjd"], np.cos(self.planet_loc[pn + "_RA"]), kind=kind
            )
            self.planet_interpolators[pn + "_RA_y"] = interp1d(
                self.planet_loc["mjd"], np.sin(self.planet_loc[pn + "_RA"]), kind=kind
            )
            self.planet_interpolators[pn + "_dec"] = interp1d(
                self.planet_loc["mjd"], self.planet_loc[pn + "_dec"], kind=kind
            )

    def get_planet_positions(self, mjd):
        result = {}
        for pn in self.planet_names:
            result[pn + "_dec"] = self.planet_interpolators[pn + "_dec"](mjd)

            temp_result = np.array(
                [
                    np.arctan2(
                        self.planet_interpolators[pn + "_RA_y"](mjd),
                        self.planet_interpolators[pn + "_RA_x"](mjd),
                    )
                ]
            ).ravel()
            negative_angles = np.where(temp_result < 0.0)[0]
            temp_result[negative_angles] = 2.0 * np.pi + temp_result[negative_angles]
            result[pn + "_RA"] = temp_result
        return result

    def get_sunset_info(self, mjd):
        """
        Returns a numpy array with mjds for various events (sunset, moonrise, sun at -12 degrees alt, etc.).
        Also the integer night number.
        """
        indx = np.searchsorted(self.sunsets["sunset"], mjd) - 1
        return self.sunsets[indx]

    def mjd_indx(self, mjd):
        indx = np.searchsorted(self.sunsets["sunset"], mjd) - 1
        return indx

    def get_sun_moon_positions(self, mjd):
        """
        All angles in Radians. moonPhase between 0 and 100.
        """
        simple_calls = ["sun_alt", "sun_dec", "moon_alt", "moon_dec", "moon_phase"]
        result = {}
        for key in simple_calls:
            result[key] = self.interpolators[key](mjd)

        longitude_calls = ["sun_az", "moon_az", "sun_RA", "moon_RA"]
        for key in longitude_calls:
            # Need to wrap in case sent a scalar
            temp_result = np.array(
                [
                    np.arctan2(
                        self.interpolators[key + "_y"](mjd),
                        self.interpolators[key + "_x"](mjd),
                    )
                ]
            ).ravel()
            negative_angles = np.where(temp_result < 0.0)[0]
            temp_result[negative_angles] = 2.0 * np.pi + temp_result[negative_angles]
            result[key] = temp_result

        return result



__all__ = ("SkyModelPre", "interp_angle")

import glob
import os
import warnings

import h5py
import healpy as hp
import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import _angular_separation, _hpid2_ra_dec, survey_start_mjd


def short_angle_dist(a0, a1):
    """
    from https://gist.github.com/shaunlebron/8832585
    """
    max_angle = 2.0 * np.pi
    da = (a1 - a0) % max_angle
    return 2.0 * da % max_angle - da


def interp_angle(x_out, xp, anglep, degrees=False):
    """
    Interpolate angle values (handle wrap around properly). Does nearest neighbor
    interpolation if values out of range.

    Parameters
    ----------
    x_out : `float` or array
        The points to interpolate to.
    xp : array
        Points to interpolate between (must be sorted)
    anglep : array
        The angles ascociated with xp
    degrees : `bool` (False)
        Set if anglep is degrees (True) or radidian (False)
    """

    # Where are the interpolation points
    x = np.atleast_1d(x_out)
    left = np.searchsorted(xp, x) - 1
    right = left + 1

    # If we are out of bounds, just use the edges
    right[np.where(right >= xp.size)] -= 1
    left[np.where(left < 0)] += 1
    baseline = xp[right] - xp[left]

    wterm = (x - xp[left]) / baseline
    wterm[np.where(baseline == 0)] = 0
    if degrees:
        result = (
            np.radians(anglep[left])
            + short_angle_dist(np.radians(anglep[left]), np.radians(anglep[right])) * wterm
        )
        result = result % (2.0 * np.pi)
        result = np.degrees(result)
    else:
        result = anglep[left] + short_angle_dist(anglep[left], anglep[right]) * wterm
        result = result % (2.0 * np.pi)
    return result


class SkyModelPre:
    """
    Load pre-computed sky brighntess maps for the LSST site and use them to interpolate to
    arbitrary dates.
    """

    def __init__(
        self,
        data_path=None,
        init_load_length=10,
        load_length=365,
        verbose=False,
        mjd0=None,
    ):
        """
        Parameters
        ----------
        data_path : `str`, opt
            path to the numpy save files. Looks in standard SIMS_SKYBRIGHTNESS_DATA or RUBIN_SIM_DATA_DIR
            if set to default (None).
        init_load_length : `int` (10)
            The length of time (days) to load from disk initially. Set to something small for fast reads.
        load_length : `int` (365)
            The number of days to load after the initial load.
        mjd0 : `float` (None)
            The starting MJD to load on initilization (days). Uses util to lookup default if None.
        """

        self.info = None
        self.sb = None
        self.header = None
        self.filter_names = None
        self.verbose = verbose

        # Look in default location for .npz files to load
        if "SIMS_SKYBRIGHTNESS_DATA" in os.environ:
            data_path = os.environ["SIMS_SKYBRIGHTNESS_DATA"]
        else:
            data_path = os.path.join(get_data_dir(), "skybrightness_pre")

        # Expect filenames of the form mjd1_mjd2.npz, e.g., 59632.155_59633.2.npz
        self.files = glob.glob(os.path.join(data_path, "*.h5"))
        if len(self.files) == 0:
            errmssg = "Failed to find pre-computed .h5 files. "
            errmssg += "Copy data from NCSA with sims_skybrightness_pre/data/data_down.sh \n"
            errmssg += "or build by running sims_skybrightness_pre/data/generate_hdf5.py"
            warnings.warn(errmssg)
        self.filesizes = np.array([os.path.getsize(filename) for filename in self.files])
        mjd_left = []
        mjd_right = []
        # glob does not always order things I guess?
        self.files.sort()
        for filename in self.files:
            temp = os.path.split(filename)[-1].replace(".h5", "").split("_")
            mjd_left.append(float(temp[0]))
            mjd_right.append(float(temp[1]))

        self.mjd_left = np.array(mjd_left)
        self.mjd_right = np.array(mjd_right)

        # Set that nothing is loaded at this point
        self.loaded_range = np.array([-1])
        self.timestep_max = -1

        if mjd0 is None:
            mjd0 = survey_start_mjd()

        # Do a quick initial load if set
        if init_load_length is not None:
            self.load_length = init_load_length
            self._load_data(mjd0)
        # swap back to the longer load length
        self.load_length = load_length
        self.nside = 32
        hpid = np.arange(hp.nside2npix(self.nside))
        self.ra, self.dec = _hpid2_ra_dec(self.nside, hpid)

    def _load_data(self, mjd, filename=None, npyfile=None):
        """
        Load up the .npz file to interpolate things. After python 3 upgrade, numpy.savez refused
        to write large .npz files, so data is split between .npz and .npy files.

        Parameters
        ----------
        mjd : `float`
            The Modified Julian Date that we want to load
        filename : `str` (None)
            The filename to restore. If None, it checks the filenames on disk to find one that
            should have the requested MJD
        npyfile : `str` (None)
            If sky brightness data not in npz file, checks the .npy file with same root name.
        """

        if filename is None:
            # Figure out which file to load.
            file_indx = np.where((mjd >= self.mjd_left) & (mjd <= self.mjd_right))[0]
            if np.size(file_indx) == 0:
                raise ValueError(
                    "MJD = %f is out of range for the files found (%f-%f)"
                    % (mjd, self.mjd_left.min(), self.mjd_right.max())
                )
            # Just take the later one, assuming we're probably simulating forward in time
            file_indx = np.max(file_indx)

            filename = self.files[file_indx]
        else:
            self.loaded_range = None

        # Use three separate try/excepet blocks so that if any of them throw
        # exceptions, we still get the others.
        try:
            del self.sb
        except AttributeError:
            pass

        try:
            del self.filter_names
        except AttributeError:
            pass

        try:
            del self.timestep_max
        except AttributeError:
            pass

        if self.verbose:
            print("Loading file %s" % filename)
        h5 = h5py.File(filename, "r")
        mjds = h5["mjds"][:]
        indxs = np.where((mjds >= mjd) & (mjds <= (mjd + self.load_length)))
        indxs = [np.min(indxs), np.max(indxs)]
        if indxs[0] > 0:
            indxs[0] -= 1
        self.loaded_range = np.array([mjds[indxs[0]], mjds[indxs[1]]])
        self.mjds = mjds[indxs[0] : indxs[1]]
        _timestep_max = np.empty(1, dtype=float)
        h5["timestep_max"].read_direct(_timestep_max)
        self.timestep_max = np.max(_timestep_max)

        self.sb = h5["sky_mags"][indxs[0] : indxs[1]]
        self.filter_names = self.sb.dtype.names
        h5.close()

        if self.verbose:
            print("%s loaded" % os.path.split(filename)[1])

        self.nside = hp.npix2nside(self.sb[self.filter_names[0]][0, :].size)

    def return_mags(
        self,
        mjd,
        indx=None,
        badval=hp.UNSEEN,
        filters=["u", "g", "r", "i", "z", "y"],
        extrapolate=False,
    ):
        """
        Return a full sky map or individual pixels for the input mjd

        Parameters
        ----------
        mjd : `float`
            Modified Julian Date to interpolate to
        indx : `List` of `int` (None)
            indices to interpolate the sky values at. Returns full sky if None. If the class was
            instatiated with opsimFields, indx is the field ID, otherwise it is the healpix ID.
        airmass_mask : `bool` (True)
            Set high (>2.5) airmass pixels to badval.
        planet_mask : `bool` (True)
            Set sky maps to badval near (2 degrees) bright planets.
        moon_mask : `bool` (True)
            Set sky maps near (10 degrees) the moon to badval.
        zenith_mask : `bool` (True)
            Set sky maps at high altitude (>86.5) to badval.
        badval : `float` (-1.6375e30)
            Mask value. Defaults to the healpy mask value.
        filters : `list`, opt
            List of strings for the filters that should be returned.
            Default returns ugrizy.
        extrapolate : `bool` (False)
            In indx is set, extrapolate any masked pixels to be the same as the nearest non-masked
            value from the full sky map.

        Returns
        -------
        sbs : `dict`
            A dictionary with filter names as keys and np.arrays as values which
            hold the sky brightness maps in mag/sq arcsec.
        """
        if mjd < self.loaded_range.min() or (mjd > self.loaded_range.max()):
            self._load_data(mjd)

        left = np.searchsorted(self.mjds, mjd) - 1
        right = left + 1

        # Do full sky by default
        if indx is None:
            indx = np.arange(self.sb["r"].shape[1])
            full_sky = True
        else:
            full_sky = False

        # If we are out of bounds
        if right >= self.mjds.size:
            right -= 1
            baseline = 1.0
        elif left < 0:
            left += 1
            baseline = 1.0
        else:
            baseline = self.mjds[right] - self.mjds[left]

        # Check if we are between sunrise/set
        if baseline > self.timestep_max + 1e-6:
            warnings.warn("Requested MJD between sunrise and sunset, returning closest maps")
            diff = np.abs(self.mjds[left.max() : right.max() + 1] - mjd)
            closest_indx = np.array([left, right])[np.where(diff == np.min(diff))].min()
            sbs = {}
            for filter_name in filters:
                sbs[filter_name] = self.sb[filter_name][closest_indx, indx]
                sbs[filter_name][np.isinf(sbs[filter_name])] = badval
                sbs[filter_name][np.where(sbs[filter_name] == hp.UNSEEN)] = badval
        else:
            wterm = (mjd - self.mjds[left]) / baseline
            w1 = 1.0 - wterm
            w2 = wterm
            sbs = {}
            for filter_name in filters:
                sbs[filter_name] = (
                    self.sb[filter_name][left, indx] * w1 + self.sb[filter_name][right, indx] * w2
                )
        # If requested a certain pixel(s), and want to extrapolate.
        if (not full_sky) & extrapolate:
            masked_pix = False
            for filter_name in filters:
                if (badval in sbs[filter_name]) | (True in np.isnan(sbs[filter_name])):
                    masked_pix = True
            if masked_pix:
                # We have pixels that are masked that we want reasonable values for
                full_sky_sb = self.return_mags(
                    mjd,
                    filters=filters,
                )
                good = np.where((full_sky_sb[filters[0]] != badval) & ~np.isnan(full_sky_sb[filters[0]]))[0]
                ra_full = self.ra[good]  # np.radians(self.header["ra"][good])
                dec_full = self.dec[good]  # np.radians(self.header["dec"][good])
                for filtername in filters:
                    full_sky_sb[filtername] = full_sky_sb[filtername][good]
                # Going to assume the masked pixels are the same in all filters
                masked_indx = np.where(
                    (sbs[filters[0]].ravel() == badval) | np.isnan(sbs[filters[0]].ravel())
                )[0]
                for i, mi in enumerate(masked_indx):
                    # Note, this is going to be really slow for many pixels, should use a kdtree
                    dist = _angular_separation(
                        self.ra[indx[i]],
                        self.dec[indx[i]],
                        ra_full,
                        dec_full,
                    )
                    closest = np.where(dist == dist.min())[0]
                    for filtername in filters:
                        sbs[filtername].ravel()[mi] = np.min(full_sky_sb[filtername][closest])

        return sbs
